import asyncio
import json
import logging
from dataclasses import dataclass, field
from typing import Optional, override, Callable, List, Awaitable, Set

from busline.client.subscriber.subscriber import Subscriber
from busline.client.subscriber.event_handler.event_handler import EventHandler
from busline.event.event import RegistryPassthroughEvent, Event, registry_passthrough_event_json_deserializer
from busline.mqtt import _MqttClientWrapper





@dataclass(kw_only=True)
class MqttSubscriber(Subscriber, _MqttClientWrapper):
    """
    Subscriber topic-based which works with MQTT

    Author: Nicola Ricciardi
    """


    deserializer: Callable[[bytes], RegistryPassthroughEvent] = field(default_factory=lambda: registry_passthrough_event_json_deserializer)
    _handle_messages_task: asyncio.Task = field(default=None, init=False)

    __subscribed_topics: Set[str] = field(default_factory=set, init=False)

    @override
    async def connect(self):
        await super().connect()
        self._handle_messages_task = asyncio.create_task(
            self._messages_handler()
        )

    async def _messages_handler(self):
        try:
            async for message in self._internal_client.messages:
                await self.notify(
                    str(message.topic),
                    self.deserializer(message.payload).to_event()
                )
        except Exception as e:
            logging.error(f"{self}: messages handler error: {repr(e)}")


    @override
    async def _internal_subscribe(self, topic: str, handler: Optional[EventHandler | Callable[[str, Event], Awaitable]] = None, **kwargs):
        await self._internal_client.subscribe(topic)
        self.__subscribed_topics.add(topic)


    @override
    async def _internal_unsubscribe(self, topic: Optional[str] = None, **kwargs):

        if topic is not None:
            await self._internal_client.unsubscribe(topic)
            self.__subscribed_topics.remove(topic)
        else:
            tasks = [
                    asyncio.create_task(self._internal_client.unsubscribe(t))
                    for t in self.__subscribed_topics
                ]

            self.__subscribed_topics.clear()

            await asyncio.gather(*tasks)



import json
from dataclasses import dataclass
from typing import Self, Tuple, Optional

from busline.event.message.avro_message import AVRO_FORMAT_TYPE, AvroMessageMixin
from busline.event.message.json_message import JsonMessageMixin, JSON_FORMAT_TYPE
from busline.event.message.message import Message
from busline.utils.serde import SerdableMixin


STRING_FORMAT_TYPE = "utf-8"


@dataclass(frozen=True)
class StringMessage(AvroMessageMixin, JsonMessageMixin, SerdableMixin):
    """
    Wrap `str` and serialize into UTF-8

    Author: Nicola Ricciardi
    """

    value: str

    def serialize(self, *, format_type: Optional[str] = None) -> Tuple[str, bytes]:

        if format_type is None or format_type == STRING_FORMAT_TYPE:
            return STRING_FORMAT_TYPE, self.value.encode("utf-8")

        if format_type == AVRO_FORMAT_TYPE:
            return AvroMessageMixin.serialize(self)

        if format_type == JSON_FORMAT_TYPE:
            return JsonMessageMixin.serialize(self)

        raise ValueError("Not supported format type")

    @classmethod
    def deserialize(cls, format_type: str, serialized_data: bytes) -> Self:
        if format_type == STRING_FORMAT_TYPE:
            return cls(serialized_data.decode("utf-8"))

        if format_type == AVRO_FORMAT_TYPE:
            return AvroMessageMixin.deserialize.__func__(cls, format_type, serialized_data)

        if format_type == JSON_FORMAT_TYPE:
            return JsonMessageMixin.deserialize.__func__(cls, format_type, serialized_data)

        raise ValueError("Not supported format type")


    @classmethod
    def from_json(cls, json_str: str) -> Self:
        return cls(json.loads(json_str)["value"])

    def to_json(self) -> str:
        return json.dumps({
            "value": self.value
        })


from typing import Dict, Type, Optional

from busline.event.message.message import Message
from busline.utils.singleton import Singleton


class EventRegistry(metaclass=Singleton):
    """
    Registry to manage different message types

    Author: Nicola Ricciardi
    """

    __associations: Dict[str, Type[Message]] = {}

    @property
    def associations(self) -> Dict[str, Type[Message]]:
        return self.__associations

    @classmethod
    def class_to_type(cls, message_class: Type[Message]) -> str:
        return message_class.__name__

    @classmethod
    def obj_to_type(cls, message: Message) -> str:
        return type(message).__name__

    def remove(self, *, message_type: Optional[str] = None, message_class: Optional[Type[Message]] = None):
        """
        Remove a message type association

        ValueError is raised if neither message_type nor message_class is provided
        """

        if message_class is None and message_type is None:
            raise ValueError("Neither message_type nor message_class is provided")

        if message_type is None:
            message_type = EventRegistry.class_to_type(message_class)

        self.__associations.pop(message_type)

    def add(self, message_class: Type[Message], *, message_type: Optional[str] = None) -> str:
        """
        Add a new association between an event message and message class.

        Return message type
        """

        if message_type is None:
            message_type = EventRegistry.class_to_type(message_class)

        self.__associations[message_type] = message_class

        return message_type

    def retrieve_class(self, message_type: str) -> Type[Message]:
        """
        Retrieve message class

        KeyError is raised if no association is found.
        """

        return self.__associations[message_type]



def add_to_registry(cls: Type[Message]):
    """
    Add to registry the decorated class

    Author: Nicola Ricciardi
    """

    # add event message in registry
    reg = EventRegistry()
    reg.add(cls)

    return cls






import logging
from typing import override, Callable
from busline.client.publisher.publisher import Publisher
from busline.event.event import Event, RegistryPassthroughEvent, registry_passthrough_event_json_serializer
from dataclasses import dataclass, field

from busline.mqtt import _MqttClientWrapper


@dataclass(kw_only=True)
class MqttPublisher(Publisher, _MqttClientWrapper):
    """
    Publisher which works with MQTT

    Author: Nicola Ricciardi
    """

    serializer: Callable[[RegistryPassthroughEvent], bytes] = field(default_factory=lambda: registry_passthrough_event_json_serializer)


    @override
    async def _internal_publish(self, topic_name: str, event: Event, **kwargs):
        await self._internal_client.publish(
            topic=topic_name,
            payload=self.serializer(RegistryPassthroughEvent.from_event(event)),
            **kwargs
        )



from abc import ABC
from typing import Self, Tuple, Optional
from dataclasses_avroschema import AvroModel
from busline.event.message.message import Message
from busline.utils.serde import SerdableMixin


AVRO_FORMAT_TYPE = "avro"


class AvroMessageMixin(Message, SerdableMixin, AvroModel, ABC):
    """
    Avro implementation for serialize/deserialize

    Author: Nicola Ricciardi
    """

    def serialize(self, *, format_type: Optional[str] = None) -> Tuple[str, bytes]:
        if format_type is not None and format_type != AVRO_FORMAT_TYPE:
            raise ValueError(f"{format_type} != {AVRO_FORMAT_TYPE}")

        return AVRO_FORMAT_TYPE, AvroModel.serialize(self, serialization_type=AVRO_FORMAT_TYPE)

    @classmethod
    def deserialize(cls, format_type: str, serialized_data: bytes) -> Self:
        if format_type != AVRO_FORMAT_TYPE:
            raise ValueError(f"{format_type} != {AVRO_FORMAT_TYPE}")

        return AvroModel.deserialize.__func__(cls, serialized_data, serialization_type=AVRO_FORMAT_TYPE)
        # __func__ instead of __call__ because `AvroModel.deserialize` is a classmethod


import asyncio
import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import List, override, Optional

from busline.event.event import Event, RegistryPassthroughEvent
from busline.client.eventbus_connector import EventBusConnector
from busline.event.message.message import Message
from busline.event.message.number_message import Int64Message, Float64Message
from busline.event.message.string_message import StringMessage


class PublishMixin(ABC):
    """
    Mixin which provides base methods to publish a message

    Author: Nicola Ricciardi
    """

    @abstractmethod
    async def publish(self, topic: str, message: Optional[Message | str | int | float] = None, **kwargs):
        raise NotImplemented()

    async def multi_publish(self, topics: List[str], message: Optional[Message | str | int | float] = None, *, parallelize: bool = True, **kwargs):
        """
        Publish the same event in more topics
        """

        logging.debug(f"{self}: publish message {message} in {len(topics)} topics (parallelization: {parallelize})")

        if parallelize:
            tasks = [self.publish(topic, message, **kwargs) for topic in topics]
            await asyncio.gather(*tasks)

        else:
            for topic in topics:
                await self.publish(topic, message, **kwargs)



@dataclass(kw_only=True, eq=False)
class Publisher(EventBusConnector, PublishMixin, ABC):
    """
    Abstract class which can be implemented by your components which must be able to publish messages

    Author: Nicola Ricciardi
    """

    def __str__(self) -> str:
        return f"Publisher('{self.identifier}')"

    @abstractmethod
    async def _internal_publish(self, topic: str, event: Event, **kwargs):
        """
        Actual publish on topic the event
        """

    @override
    async def publish(self, topic: str, message: Optional[Message | str | int | float] = None,
                      *, event_timestamp: Optional[datetime] = None, event_identifier: Optional[str] = None, **kwargs) -> Event:
        """
        Publish on topic the message and return the generated event.

        If str, int or float are used, the message is wrapper in StringMessage, Int64Message or Float64Message.

        Event use datetime.now() for timestamp and uuid4() for identifier by default.
        """

        if message is not None:
            if isinstance(message, str):
                message = StringMessage(message)

            if isinstance(message, int):
                message = Int64Message(message)

            if isinstance(message, float):
                message = Float64Message(message)

        if event_timestamp is None:
            event_identifier = datetime.now()

        if event_identifier is None:
            event_identifier = str(uuid.uuid4())

        event = Event(
            payload=message,
            identifier=event_identifier,
            timestamp=event_identifier,
            publisher_identifier=self.identifier
        )

        logging.info(f"{self}: publish on {topic} -> {event}")
        await self._on_publishing(topic, event, **kwargs)
        await self._internal_publish(topic, event, **kwargs)
        await self._on_published(topic, event, **kwargs)

        return event


    async def _on_publishing(self, topic: str, event: Event, **kwargs):
        """
        Callback called on publishing start
        """

    async def _on_published(self, topic: str, event: Event, **kwargs):
        """
        Callback called on publishing end
        """


from abc import ABC, abstractmethod
from dataclasses import dataclass, asdict
from typing import Self, Tuple

from dataclasses_avroschema import AvroModel
import json

from busline.utils.serde import SerdableMixin


BYTES_FORMAT_TYPE = "bytes"


class Message(SerdableMixin, ABC):
    """
    Sendable message, it must be serializable and deserializable

    Author: Nicola Ricciardi
    """





import inspect
from typing import TypeVar, Callable, Coroutine, Dict, List

from busline.client.subscriber.event_handler.callback_event_handler import CallbackEventHandler
from busline.event.event import Event


F = TypeVar('F', bound=Callable[[str, Event], Coroutine])


def event_handler(func):
    """
    Transform a function or a method into an event handler

    Author: Nicola Ricciardi
    """

    def decorator(method_or_func):
        is_method = 'self' in method_or_func.__code__.co_varnames

        if not is_method:
            # standalone async function
            return CallbackEventHandler(on_event_callback=method_or_func)

        # method in a class
        method_name = method_or_func.__name__

        class HandlerDescriptor:
            def __get__(self, instance, owner):
                if instance is None:
                    return self

                # callback that binds the method to self
                async def callback(topic: str, event: Event):
                    await method_or_func(instance, topic, event)

                handler = CallbackEventHandler(on_event_callback=callback)

                # replace the method with the handler
                setattr(instance, method_name, handler)
                return handler

        return HandlerDescriptor()

    return decorator(func)








from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set
from busline.exceptions import TopicNotFound
from busline.client.subscriber.subscriber import Subscriber
from busline.event.event import Event


@dataclass
class EventBus(ABC):
    """
    Abstract class used as base for new eventbus implemented in local projects.

    Author: Nicola Ricciardi
    """

    subscriptions: Dict[str, Set[Subscriber]] = field(default_factory=lambda: defaultdict(set), init=False)

    @property
    def topics(self) -> List[str]:
        return list(self.subscriptions.keys())

    def reset_subscriptions(self):
        self.subscriptions = defaultdict(set)

    def add_subscriber(self, topic: str, subscriber: Subscriber):
        """
        Add subscriber to topic

        :param topic:
        :param subscriber:
        :return:
        """

        self.subscriptions[topic].add(subscriber)

    def remove_subscriber(self, subscriber: Subscriber, topic: Optional[str] = None, raise_if_topic_missed: bool = False):
        """
        Remove subscriber from topic selected or from all if topic is None

        :param raise_if_topic_missed:
        :param subscriber:
        :param topic:
        :return:
        """

        if raise_if_topic_missed and topic is not None and topic not in self.subscriptions.keys():
            raise TopicNotFound(f"topic '{topic}' not found")

        for name in self.subscriptions.keys():

            if topic is None or self._topic_names_match(topic, name):
                if subscriber in self.subscriptions[name]:
                    self.subscriptions[name].remove(subscriber)


    def _topic_names_match(self, t1: str, t2: str):
        return t1 == t2

    def _get_topic_subscriptions(self, topic: str) -> Set[Subscriber]:

        topic_subscriptions: Set[Subscriber] = set()
        for t, subs in self.subscriptions.items():
            if self._topic_names_match(t, topic):
                topic_subscriptions = topic_subscriptions.union(subs)

        return topic_subscriptions

    @abstractmethod
    async def put_event(self, topic: str, event: Event):
        """
        Put a new event in the bus and notify subscribers of corresponding
        event's topic

        :param topic:
        :param event:
        :return:
        """

        raise NotImplemented()




import json
from dataclasses import dataclass, field
from typing import Self, Tuple, Literal, Optional
import struct

from busline.event.message.avro_message import AVRO_FORMAT_TYPE, AvroMessageMixin
from busline.event.message.json_message import JSON_FORMAT_TYPE, JsonMessageMixin
from busline.event.message.message import Message, BYTES_FORMAT_TYPE
from busline.utils.serde import SerdableMixin




@dataclass(frozen=True)
class Int64Message(AvroMessageMixin, JsonMessageMixin, SerdableMixin):
    """
    Wrap `int` and serialize into 8 bytes integer number

    Author: Nicola Ricciardi
    """

    value: int

    def serialize(self, *, format_type: Optional[str] = None) -> Tuple[str, bytes]:
        if format_type == BYTES_FORMAT_TYPE:
            return BYTES_FORMAT_TYPE, self.value.to_bytes(length=8, signed=True, byteorder="big")

        if format_type == AVRO_FORMAT_TYPE:
            return AvroMessageMixin.serialize(self)

        if format_type == JSON_FORMAT_TYPE:
            return JsonMessageMixin.serialize(self)

        raise ValueError("Not supported format type")

    @classmethod
    def deserialize(cls, format_type: str, serialized_data: bytes) -> Self:
        if format_type is None or format_type == BYTES_FORMAT_TYPE:
            return cls(int.from_bytes(serialized_data, signed=True, byteorder="big"))

        if format_type == AVRO_FORMAT_TYPE:
            return AvroMessageMixin.deserialize.__func__(cls, format_type, serialized_data)

        if format_type == JSON_FORMAT_TYPE:
            return JsonMessageMixin.deserialize.__func__(cls, format_type, serialized_data)

        raise ValueError("Not supported format type")

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        return cls(json.loads(json_str)["value"])

    def to_json(self) -> str:
        return json.dumps({
            "value": self.value
        })



@dataclass(frozen=True)
class Int32Message(AvroMessageMixin, JsonMessageMixin, SerdableMixin):
    """
    Wrap `int` and serialize into 4 bytes integer number

    Author: Nicola Ricciardi
    """

    value: int

    def serialize(self, *, format_type: Optional[str] = None) -> Tuple[str, bytes]:
        if format_type == BYTES_FORMAT_TYPE:
            return BYTES_FORMAT_TYPE, self.value.to_bytes(length=4, byteorder="big")

        if format_type == AVRO_FORMAT_TYPE:
            return AvroMessageMixin.serialize(self)

        if format_type == JSON_FORMAT_TYPE:
            return JsonMessageMixin.serialize(self)

        raise ValueError("Not supported format type")

    @classmethod
    def deserialize(cls, format_type: str, serialized_data: bytes) -> Self:
        if format_type is None or format_type == BYTES_FORMAT_TYPE:
            return cls(int.from_bytes(serialized_data, byteorder="big"))

        if format_type == AVRO_FORMAT_TYPE:
            return AvroMessageMixin.deserialize.__func__(cls, format_type, serialized_data)

        if format_type == JSON_FORMAT_TYPE:
            return JsonMessageMixin.deserialize.__func__(cls, format_type, serialized_data)

        raise ValueError("Not supported format type")

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        return cls(json.loads(json_str)["value"])

    def to_json(self) -> str:
        return json.dumps({
            "value": self.value
        })


@dataclass(frozen=True)
class Float32Message(AvroMessageMixin, JsonMessageMixin, SerdableMixin):
    """
    Wrap `float` and serialize into 4 bytes floating point

    Author: Nicola Ricciardi
    """

    value: float

    def serialize(self, *, format_type: Optional[str] = None) -> Tuple[str, bytes]:
        if format_type is None or format_type == BYTES_FORMAT_TYPE:
            return BYTES_FORMAT_TYPE, struct.pack(">f", self.value)

        if format_type == AVRO_FORMAT_TYPE:
            return AvroMessageMixin.serialize(self)

        if format_type == JSON_FORMAT_TYPE:
            return JsonMessageMixin.serialize(self)

        raise ValueError("Not supported format type")

    @classmethod
    def deserialize(cls, format_type: str, serialized_data: bytes) -> Self:
        if format_type == BYTES_FORMAT_TYPE:
            return cls(struct.unpack(">f", serialized_data)[0])

        if format_type == AVRO_FORMAT_TYPE:
            return AvroMessageMixin.deserialize.__func__(cls, format_type, serialized_data)

        if format_type == JSON_FORMAT_TYPE:
            return JsonMessageMixin.deserialize.__func__(cls, format_type, serialized_data)

        raise ValueError("Not supported format type")

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        return cls(json.loads(json_str)["value"])

    def to_json(self) -> str:
        return json.dumps({
            "value": self.value
        })


@dataclass(frozen=True)
class Float64Message(AvroMessageMixin, JsonMessageMixin, SerdableMixin):
    """
    Wrap `float` and serialize into 8 bytes floating point

    Author: Nicola Ricciardi
    """

    value: float

    def serialize(self, *, format_type: Optional[str] = None) -> Tuple[str, bytes]:
        if format_type is None or format_type == BYTES_FORMAT_TYPE:
            return BYTES_FORMAT_TYPE, struct.pack(">d", self.value)

        if format_type == AVRO_FORMAT_TYPE:
            return AvroMessageMixin.serialize(self)

        if format_type == JSON_FORMAT_TYPE:
            return JsonMessageMixin.serialize(self)

        raise ValueError("Not supported format type")

    @classmethod
    def deserialize(cls, format_type: str, serialized_data: bytes) -> Self:
        if format_type == BYTES_FORMAT_TYPE:
            return cls(struct.unpack(">d", serialized_data)[0])

        if format_type == AVRO_FORMAT_TYPE:
            return AvroMessageMixin.deserialize.__func__(cls, format_type, serialized_data)

        if format_type == JSON_FORMAT_TYPE:
            return JsonMessageMixin.deserialize.__func__(cls, format_type, serialized_data)

        raise ValueError("Not supported format type")

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        return cls(json.loads(json_str)["value"])

    def to_json(self) -> str:
        return json.dumps({
            "value": self.value
        })



import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import Queue
from dataclasses import dataclass, field
from idlelib.window import add_windows_to_menu
from typing import Optional, override, List, Tuple, Dict, Callable, Any, AsyncGenerator, Awaitable

from busline.client.eventbus_connector import EventBusConnector
from busline.client.subscriber.event_handler import CallbackEventHandler, event_handler
from busline.client.subscriber.event_handler.event_handler import EventHandler
from busline.event.event import Event
from busline.exceptions import EventHandlerNotFound


class SubscribeMixin(ABC):
    """
    Mixin which provides base methods to subscribe and unsubscribe

    Author: Nicola Ricciardi
    """

    @abstractmethod
    async def subscribe(self, topic: str, **kwargs):
        raise NotImplemented()

    async def multi_subscribe(self, topics: List[str], /, parallelize: bool = True, **kwargs):
        logging.debug(f"{self}: subscribe to {len(topics)} topics (parallelization: {parallelize})")

        if parallelize:
            tasks = [self.subscribe(topic, **kwargs) for topic in topics]
            await asyncio.gather(*tasks)

        else:
            for topic in topics:
                await self.subscribe(topic, **kwargs)

    @abstractmethod
    async def unsubscribe(self, topic: Optional[str] = None, **kwargs):
        raise NotImplemented()

    async def multi_unsubscribe(self, topics: List[str], /, parallelize: bool = True, **kwargs):
        logging.debug(f"{self}: unsubscribe from {len(topics)} topics (parallelization: {parallelize})")

        if parallelize:
            tasks = [self.unsubscribe(topic, **kwargs) for topic in topics]
            await asyncio.gather(*tasks)

        else:
            for topic in topics:
                await self.unsubscribe(topic, **kwargs)



@dataclass(kw_only=True, eq=False)
class Subscriber(EventBusConnector, SubscribeMixin, ABC):
    """
    Handles different topic events using ad hoc handlers defined by user,
    else it uses fallback handler if provided (otherwise throws an exception)

    Attributes:
        default_handler: event handler used for a topic if no event handler is specified for that topic
        topic_names_matcher: function used to check match between two topic name (with wildcards); default "t1 == t2"
        handler_always_required: raise an exception if no handlers are found for a topic

    Author: Nicola Ricciardi
    """

    default_handler: Optional[EventHandler] = field(default=None)
    topic_names_matcher: Callable[[str, str], bool] = field(repr=False, default=lambda t1, t2: t1 == t2)
    handler_always_required: bool = field(default=False)
    _handlers: Dict[str, EventHandler] = field(default_factory=dict, init=False)
    _inbound_events_queue: Queue[Tuple[str, Event]] = field(default_factory=lambda: Queue(maxsize=0), init=False)
    _inbound_not_handled_events_queue: Queue[Tuple[str, Event]] = field(default_factory=lambda: Queue(maxsize=0), init=False)
    _stop_queue_processing: asyncio.Event = field(default_factory=lambda: asyncio.Event(), init=False)

    @override
    async def connect(self):
        await super().connect()
        self._stop_queue_processing.clear()

    @override
    async def disconnect(self):
        await super().disconnect()
        self._stop_queue_processing.set()

    @abstractmethod
    async def _internal_subscribe(self, topic: str, handler: Optional[EventHandler] = None, **kwargs):
        """
        Actual subscribe to topic
        """

    @abstractmethod
    async def _internal_unsubscribe(self, topic: Optional[str] = None, **kwargs):
        """
        Actual unsubscribe to topic
        """

    @override
    async def subscribe(self, topic: str, handler: Optional[EventHandler | Callable[[str, Event], Awaitable]] = None, **kwargs):
        """
        Subscribe to topic using handler. If handler is an async callback, then it is wrapped using CallbackEventHandler.
        Otherwise, if it is None, then default handler will be used (if it will be set).
        """

        if handler is not None:
            if not issubclass(type(handler), EventHandler):
                handler = event_handler(handler)

        logging.info(f"{self}: subscribe on topic {topic}")
        await self._on_subscribing(topic, handler, **kwargs)
        await self._internal_subscribe(topic, handler, **kwargs)
        await self._on_subscribed(topic, handler, **kwargs)

    @override
    async def unsubscribe(self, topic: Optional[str] = None, **kwargs):
        """
        Unsubscribe to topic
        """

        logging.info(f"{self}: unsubscribe from topic {topic}")
        await self._on_unsubscribing(topic, **kwargs)
        await self._internal_unsubscribe(topic, **kwargs)
        await self._on_unsubscribed(topic, **kwargs)

    async def _on_subscribing(self, topic: str, handler: Optional[EventHandler] = None, **kwargs):
        """
        Callback called on subscribing
        """

    async def _on_subscribed(self, topic: str, handler: Optional[EventHandler] = None, **kwargs):
        """
        Callback called on subscribed
        """

        self._handlers[topic] = handler

    async def _on_unsubscribing(self, topic: Optional[str], **kwargs):
        """
        Callback called on unsubscribing
        """

    async def _on_unsubscribed(self, topic: Optional[str], **kwargs):
        """
        Callback called on unsubscribed
        """

        if topic is None:
            self._handlers = {}
        else:
            if topic in self._handlers:
                del self._handlers[topic]
            else:
                logging.warning(f"{self}: unsubscribed from unknown topic: {topic}")

    def __get_handlers_of_topic(self, topic: str) -> List[EventHandler]:

        handlers = []
        for t, h in self._handlers.items():
            if not self.topic_names_matcher(topic, t):
                continue

            if h is not None:
                handlers.append(h)
                continue

            if self.default_handler is not None:
                handlers.append(self.default_handler)
                continue

            if self.handler_always_required:
                raise EventHandlerNotFound()
            else:
                logging.warning(f"{self}: event handler for topic '{topic}' not found")

        return handlers


    async def notify(self, topic: str, event: Event, **kwargs):
        """
        Notify subscriber
        """

        handlers_of_topic: List[EventHandler] = self.__get_handlers_of_topic(topic)

        tasks = [
            self._inbound_events_queue.put((topic, event))
        ]

        if len(handlers_of_topic) > 0:
            tasks.extend([handler.handle(topic, event) for handler in handlers_of_topic])
        else:
            tasks.append(self._inbound_not_handled_events_queue.put((topic, event)))

        await asyncio.gather(
            *tasks
        )


    @property
    async def inbound_events(self) -> AsyncGenerator[tuple[str, Event], None]:
        """
        Generator which provides all inbound events
        """

        while not self._stop_queue_processing.is_set():
            topic, event = await self._inbound_events_queue.get()
            yield topic, event

    @property
    async def inbound_unhandled_events(self) -> AsyncGenerator[tuple[str, Event], None]:
        """
        Generator which provides only inbound events which are not handled
        """

        while not self._stop_queue_processing.is_set():
            topic, event = await self._inbound_not_handled_events_queue.get()
            yield topic, event


    def __str__(self) -> str:
        return f"Subscriber('{self.identifier}')"


from abc import ABC, abstractmethod
from busline.event.event import Event


class EventHandler(ABC):
    """

    Author: Nicola Ricciardi
    """

    @abstractmethod
    async def handle(self, topic: str, event: Event):
        """
        Manage an event of a topic
        """


import logging
from dataclasses import dataclass, field
from typing import Optional, override, Callable, Awaitable

from busline.client.subscriber.event_handler.event_handler import EventHandler
from busline.client.subscriber.subscriber import Subscriber
from busline.event.event import Event
from busline.local.eventbus.eventbus import EventBus
from busline.exceptions import EventBusClientNotConnected


@dataclass(eq=False)
class LocalSubscriber(Subscriber):
    """
    Subscriber topic-based which works with local eventbus

    Author: Nicola Ricciardi
    """

    eventbus: EventBus
    connected: bool = field(default=False)

    @override
    async def connect(self):
        logging.info(f"{self}: connecting...")
        self.connected = True

    @override
    async def disconnect(self):
        logging.info(f"{self}: disconnecting...")
        self.connected = False

    @override
    async def _internal_subscribe(self, topic: str, handler: Optional[EventHandler | Callable[[str, Event], Awaitable]] = None, **kwargs):
        if not self.connected:
            raise EventBusClientNotConnected()

        self.eventbus.add_subscriber(topic, self)

    @override
    async def _internal_unsubscribe(self, topic: Optional[str] = None, **kwargs):
        if not self.connected:
            raise EventBusClientNotConnected()

        self.eventbus.remove_subscriber(self, topic)

    def __eq__(self, other):
        return self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)


from __future__ import annotations

import base64
import json
import uuid
from datetime import datetime
from dataclasses import dataclass, field, asdict
from typing import Optional, Generic, TypeVar, Dict, Self
from abc import ABC

from dataclasses_avroschema import AvroModel

from busline.utils.serde import SerdableMixin
from busline.event.registry import EventRegistry


_registry = EventRegistry()



M = TypeVar('M', bound='Message')



@dataclass(kw_only=True)
class Event(Generic[M]):
    """
    Python-level inbound event

    Author: Nicola Ricciardi
    """

    publisher_identifier: str
    payload: Optional[M]
    identifier: str
    timestamp: datetime


@dataclass(kw_only=True)
class RegistryPassthroughEvent(AvroModel):
    """
    Utility class to manage event serialization. It works with EventRegistry to manage deserialization.

    Author: Nicola Ricciardi
    """

    identifier: str
    publisher_identifier: str
    serialized_payload: Optional[bytes]
    payload_format_type: Optional[str]
    message_type: Optional[str]
    timestamp: datetime


    @classmethod
    def from_event(cls, event: Event[M], *, message_type: Optional[str] = None) -> Self:
        """
        Automatically add event message to registry
        """

        if message_type is None and event.payload is not None:
            message_type = _registry.add(type(event.payload), message_type=message_type)

        payload_format_type, serialized_payload = event.payload.serialize() if event.payload is not None else (None, None)

        return cls(
            publisher_identifier=event.publisher_identifier,
            timestamp=event.timestamp,
            identifier=event.identifier,
            message_type=message_type,
            serialized_payload=serialized_payload,
            payload_format_type=payload_format_type
        )

    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        return cls(
            identifier=data["identifier"],
            publisher_identifier=data["publisher_identifier"],
            serialized_payload=data["serialized_payload"],
            message_type=data["message_type"],
            timestamp=data["timestamp"],
            payload_format_type=data["payload_format_type"]
        )

    def to_event(self) -> Event:
        """
        Retrieve from registry right message class, then construct the event
        """

        payload = None
        if self.serialized_payload is not None:
            if self.message_type is None:
                raise ValueError("Message type missed")

            class_of_message = _registry.retrieve_class(self.message_type)
            payload = class_of_message.deserialize(self.payload_format_type, self.serialized_payload)

        return Event(
            identifier=self.identifier,
            publisher_identifier=self.publisher_identifier,
            timestamp=self.timestamp,
            payload=payload
        )

    def to_dict(self) -> Dict:
        return {
            "identifier": self.identifier,
            "publisher_identifier": self.publisher_identifier,
            "serialized_payload": self.serialized_payload,
            "payload_format_type": self.payload_format_type,
            "message_type": self.message_type,
            "timestamp": self.timestamp
        }


def _custom_encoder(obj):
    if isinstance(obj, datetime):
        return {"__type__": "datetime", "value": obj.isoformat()}
    if isinstance(obj, bytes):
        return {"__type__": "bytes", "value": base64.b64encode(obj).decode("utf-8")}

    return str(obj)


def registry_passthrough_event_json_serializer(event: RegistryPassthroughEvent) -> bytes:
    return json.dumps(event.to_dict(), default=_custom_encoder).encode("utf-8")


def _custom_decoder(obj):
    if "__type__" in obj:
        if obj["__type__"] == "datetime":
            return datetime.fromisoformat(obj["value"])
        if obj["__type__"] == "bytes":
            return base64.b64decode(obj["value"])
    return obj

def registry_passthrough_event_json_deserializer(serialized_event: bytes) -> RegistryPassthroughEvent:
    return RegistryPassthroughEvent.from_dict(json.loads(serialized_event.decode("utf-8"), object_hook=_custom_decoder))






import logging
from abc import ABC
from dataclasses import dataclass, field
from typing import override, Any, Dict
from aiomqtt import Client

from busline.client.eventbus_connector import EventBusConnector


@dataclass(kw_only=True)
class _MqttClientWrapper(EventBusConnector, ABC):

    hostname: str
    port: int = field(default=1883)
    other_client_parameters: Dict = field(default_factory=dict)
    _internal_client: Client = field(default=None, init=False)
    __context: Any = field(init=False)


    def __post_init__(self):
        self._internal_client = Client(
            hostname=self.hostname,
            port=self.port,
            identifier=self.identifier,
            **self.other_client_parameters
        )

    @override
    async def connect(self):
        logging.info(f"{self}: connecting...")

        if self._internal_client is None:
            raise ValueError("Internal client is not set")

        # open context, equivalent to "async with"
        self.__context = self._internal_client.__aenter__()
        await self.__context

    @override
    async def disconnect(self):
        logging.info(f"{self}: disconnecting...")

        # close context
        await self._internal_client.__aexit__(None, None, None)
        self.__context = None





from abc import ABC, abstractmethod
from typing import Tuple, Self, Optional


class SerializableMixin(ABC):
    """
    Author: Nicola Ricciardi
    """


    @abstractmethod
    def serialize(self, *, format_type: Optional[str] = None) -> Tuple[str, bytes]:
        """
        Serialize itself and return (format type, serialized data).
        For example, ("json", "{...}").

        Explicitly provide format_type to choose it, otherwise default is used
        """

        raise NotImplemented()


class DeserializableMixin(ABC):
    """
    Author: Nicola Ricciardi
    """

    @classmethod
    @abstractmethod
    def deserialize(cls, format_type: str, serialized_data: bytes) -> Self:
        raise NotImplemented()


class SerdableMixin(SerializableMixin, DeserializableMixin, ABC):
    pass




import inspect
from typing import Callable, Coroutine, Dict, override, List, Any, Awaitable
from dataclasses import dataclass, field

from busline.event.event import Event
from busline.client.subscriber.event_handler.event_handler import EventHandler


@dataclass
class CallbackEventHandler(EventHandler):
    """
    Event handler which use a pre-defined callback

    Author: Nicola Ricciardi
    """

    on_event_callback: Callable[[str, Event], Any]

    __actual_async_on_event_callback: Callable[[str, Event], Awaitable] = field(init=False)

    def __post_init__(self):
        if not inspect.iscoroutinefunction(self.on_event_callback):

            async def async_wrapper(topic: str, event: Event) -> Callable[[str, Event], Awaitable]:
                return self.on_event_callback(topic, event)

            self.__actual_async_on_event_callback = async_wrapper

        else:
            self.__actual_async_on_event_callback = self.on_event_callback

    async def handle(self, topic: str, event: Event):
        await self.__actual_async_on_event_callback(topic, event)




from abc import ABC, abstractmethod
from dataclasses import dataclass, asdict
from typing import Self, Tuple, Optional
import json
from busline.event.message.message import Message
from busline.utils.serde import SerdableMixin


JSON_FORMAT_TYPE = "json"


@dataclass(frozen=True)
class JsonMessageMixin(Message, SerdableMixin, ABC):
    """
    JSON implementation for serialize/deserialize

    Author: Nicola Ricciardi
    """

    @classmethod
    @abstractmethod
    def from_json(cls, json_str: str) -> Self:
        raise NotImplemented()

    @abstractmethod
    def to_json(self) -> str:
        raise NotImplemented()

    def serialize(self, *, format_type: Optional[str] = None) -> Tuple[str, bytes]:
        if format_type is not None and format_type != JSON_FORMAT_TYPE:
            raise ValueError(f"{format_type} != {JSON_FORMAT_TYPE}")

        return JSON_FORMAT_TYPE, self.to_json().encode("utf-8")

    @classmethod
    def deserialize(cls, format_type: str, serialized_data: bytes) -> Self:
        if format_type != JSON_FORMAT_TYPE:
            raise ValueError(f"{format_type} != {JSON_FORMAT_TYPE}")

        return cls.from_json(serialized_data.decode("utf-8"))
